#!/usr/bin/env python3


import re

from .IDCounter import IDCounter

from .EnumLogLevel import EnumLogLevel
from .AbstractLogger import AbstractLogger
from .BufferLogger import BufferLogger
from .ConsoleLogger import ConsoleLogger
from .DetectionLogger import DetectionLogger
from .FilterLogger import FilterLogger
from .MulticastLogger import MulticastLogger
from .NamedMulticastLogger import NamedMulticastLogger
from .NullLogger import NullLogger
#from .SimpleFileLogger import SimpleFileLogger
from .FileLogger import FileLogger
from .StringListLogger import StringListLogger
from .JSONLogger import JSONLogger

from .AbstractLogMessageFormatter import AbstractLogMessageFormatter
from .LogMessageFormatter import LogMessageFormatter, DEFAULT_LOG_MESSAGE_FORMATTER
from .ColoredLogMessageFormatter import ColoredLogMessageFormatter, COLOR_LOG_MESSAGE_FORMATTER





def instantiateLogMsgFormatter(cfg):
	if isinstance(cfg, dict):
		logMsgFormatterType = cfg["type"]
		if logMsgFormatterType == "default":
			if "extensitivity" in cfg:
				extensitivity = cfg["extensitivity"]
				if extensitivity == "full":
					DEFAULT_LOG_MESSAGE_FORMATTER.setOutputMode(DEFAULT_LOG_MESSAGE_FORMATTER.EnumOutputMode.FULL)
				elif extensitivity in [ "short", "shorted", "shortened" ]:
					DEFAULT_LOG_MESSAGE_FORMATTER.setOutputMode(DEFAULT_LOG_MESSAGE_FORMATTER.EnumOutputMode.SHORTED)
				elif extensitivity == "veryShort":
					DEFAULT_LOG_MESSAGE_FORMATTER.setOutputMode(DEFAULT_LOG_MESSAGE_FORMATTER.EnumOutputMode.VERY_SHORT)
				else:
					raise Exception("Unknown extensitivity: " + repr(extensitivity))
			return DEFAULT_LOG_MESSAGE_FORMATTER
		elif logMsgFormatterType == "color":
			if "extensitivity" in cfg:
				extensitivity = cfg["extensitivity"]
				if extensitivity == "full":
					COLOR_LOG_MESSAGE_FORMATTER.setOutputMode(COLOR_LOG_MESSAGE_FORMATTER.EnumOutputMode.FULL)
				elif extensitivity in [ "short", "shorted", "shortened" ]:
					COLOR_LOG_MESSAGE_FORMATTER.setOutputMode(COLOR_LOG_MESSAGE_FORMATTER.EnumOutputMode.SHORTED)
				elif extensitivity == "veryShort":
					COLOR_LOG_MESSAGE_FORMATTER.setOutputMode(COLOR_LOG_MESSAGE_FORMATTER.EnumOutputMode.VERY_SHORT)
				else:
					raise Exception("Unknown extensitivity: " + repr(extensitivity))
			return COLOR_LOG_MESSAGE_FORMATTER
		else:
			raise Exception("Unknown log message formatter: " + repr(cfg))
	else:
		if cfg == "default":
			return DEFAULT_LOG_MESSAGE_FORMATTER
		elif cfg == "color":
			return COLOR_LOG_MESSAGE_FORMATTER
		else:
			raise Exception("Unknown log message formatter: " + repr(cfg))
#



def instantiate(cfg):
	loggerType = cfg["type"]

	if loggerType == "BufferLogger":
		return BufferLogger.create()

	elif loggerType == "ConsoleLogger":
		logMsgFormatter = None
		if "logMsgFormatter" in cfg:
			logMsgFormatter = instantiateLogMsgFormatter(cfg["logMsgFormatter"])
		return ConsoleLogger.create(logMsgFormatter = logMsgFormatter)

	elif loggerType == "DetectionLogger":
		return DetectionLogger.create(instantiate(cfg["nested"]))

	elif loggerType == "FilterLogger":
		if "minLogLevel" in cfg:
			logLevel = EnumLogLevel.parse(cfg["minLogLevel"])
		else:
			logLevel = EnumLogLevel.WARNING
		return FilterLogger.create(instantiate(cfg["nested"]), minLogLevel = logLevel)

	elif loggerType == "MulticastLogger":
		loggers = []
		for item in cfg["nested"]:
			loggers.append(instantiate(item))
		return MulticastLogger.create(*loggers)

	elif loggerType == "NamedMulticastLogger":
		loggers = {}
		for itemKey in cfg["nested"]:
			loggers[itemKey] = instantiate(cfg["nested"][itemKey])
		return NamedMulticastLogger.create(**loggers)

	elif loggerType == "NullLogger":
		return NullLogger.create()

	elif loggerType == "FileLogger":
		mode = cfg.get("fileMode", None)
		if mode != None:
			if isinstance(mode, int):
				mode = str(mode)
			if isinstance(mode, str):
				if re.match("^[0-7][0-7][0-7]$", mode):
					mode = int(mode, 8)
				else:
					raise Exception("Invalid mode specified for file logger!")
			else:
				raise Exception("Invalid mode specified for file logger!")
		return FileLogger.create(
			cfg["filePath"],
			cfg.get("rollOver", None),
			cfg.get("bAppendToExistingFile", True),
			cfg.get("bFlushAfterEveryLogMessage", True),
			mode)

	elif loggerType == "StringListLogger":
		return StringListLogger.create()

	elif loggerType == "JSONLogger":
		return JSONLogger.create(cfg["filePath"])

	else:
		raise Exception("Unknown logger type: " + loggerType)
#













#!/usr/bin/env python3



import os
import time
import traceback
import sys
import abc

import sh

from .EnumLogLevel import *
from .AbstractLogger import *





#
# This logger will buffer log messages in an array.
#
class BufferLogger(AbstractLogger):



	def __init__(self, indentLevel = 0, prefix = "", logItemList = None):
		self.__level = indentLevel
		self.__prefix = prefix
		if logItemList is None:
			self.__list = []
		else:
			self.__list = logItemList



	#
	# Return a list of strings that contains the data stored in this logger.
	# Standard formatting is used for output.
	#
	# @return		string[]		Returns an array of strings ready to be written to the console or a file.
	#
	def getBufferDataAsStrList(self):
		ret = []
		for (timeStamp, logLevel, textOrException) in self.__list:
			lineOrLines = self._logEntryToStringOrStringList(timeStamp, logLevel, textOrException)
			if isinstance(lineOrLines, str):
				ret.append(self.__prefix + lineOrLines)
			else:
				for line in lineOrLines:
					ret.append(self.__prefix + line)
		return ret



	#
	# Return a list of tuples that contains the data stored in this logger.
	#
	# @return		tuple[]		Returns an array of tuples. Each tuple will contain the following fields:
	#							* int timeStamp : The time stamp since Epoch in seconds.
	#							* EnumLogLevel logLevel : The log level of this log entry.
	#							* string|Exception textOrException : A log message or an execption object.
	#
	def getBufferDataAsTupleList(self):
		return self.__list



	#
	# Return a single string that contains the data stored in this logger.
	# Standard formatting is used for output.
	#
	# @return		string		Returns a single string ready to be written to the console or a file.
	#
	def getBufferDataAsStr(self):
		s = ''
		for (timeStamp, logLevel, textOrException) in self.__list:
			lineOrLines = self._logEntryToStringOrStringList(timeStamp, logLevel, textOrException)
			if isinstance(lineOrLines, str):
				s += self.__prefix + lineOrLines + "\n"
			else:
				for line in lineOrLines:
					s += self.__prefix + line + "\n"
		return s



	def _log(self, timeStamp, logLevel, textOrException):
		self.__list.append((timeStamp, logLevel, textOrException))



	def descend(self, text):
		self._log(time.time(), EnumLogLevel.INFO, text)
		return BufferLogger(self.__level + 1, self.__prefix + "\t", self.__list)



	#
	# Forward the log data stord in this logger to another logger.
	#
	# @param		AbstractLogger logger			Another logger that will receive the log data.
	#
	def forwardTo(self, logger, bClear = False):
		assert isinstance(logger, AbstractLogger)

		logger._logAll(self.__list)

		if bClear:
			self.__list = []



	def clear(self):
		self.__list = []










#!/usr/bin/env python3



import os
import time
import traceback
import sys
import abc
import datetime

import sh

from .EnumLogLevel import *
from .AbstractLogger import *





#
# This logger will buffer log messages in an internal array. Later this data can be forwarded to
# other loggers, f.e. in order to store them on disk.
#
class BufferLogger(AbstractLogger):



	def __init__(self, idCounter = None, parentID = None, indentLevel = 0, logItemList = None):
		super().__init__(idCounter)
		self._indentationLevel = indentLevel
		if logItemList is None:
			self.__list = []
		else:
			self.__list = logItemList
		if parentID is None:
			parentID = self._idCounter.next()
		self._parentLogEntryID = parentID
	#



	@staticmethod
	def __convertRawLogData(items):
		ret = []
		for item in items:
			item = list(item)
			item[5] = EnumLogLevel.parse(item[5])
			if item[0] == "txt":
				pass
			elif item[0] == "ex":
				pass
			elif item[0] == "desc":
				item[7] = BufferLogger.__convertRawLogData(item[7])
			else:
				raise Exception("Implementation Error!")
			ret.append(item)
		return ret
	#



	@staticmethod
	def create(jsonRawData = None):
		if jsonRawData != None:
			jsonRawData = BufferLogger.__convertRawLogData(jsonRawData)
			return BufferLogger(None, None, 0, jsonRawData)
		return BufferLogger()
	#



	def hasData(self):
		return len(self.__list) > 0
	#



	"""
	#
	# Return a list of strings that contains the data stored in this logger.
	#
	# @return		string[]		Returns an array of strings ready to be written to the console or a file.
	#
	def getBufferDataAsStrList(self):
		ret = []
		for logEntryStruct in in self.__list:
			...
		return ret
	"""




	"""
	#
	# Return a single string that contains the data stored in this logger.
	#
	# @return		string		Returns a single string ready to be written to the console or a file.
	#
	def getBufferDataAsStr(self):
		s = ''
		for logEntryStruct in in self.__list:
			...
		return s
	"""



	def _logi(self, logEntryStruct, bNeedsIndentationLevelAdaption):
		if bNeedsIndentationLevelAdaption:
			logEntryStruct = list(logEntryStruct)
			logEntryStruct[2] = self._indentationLevel
		self.__list.append(logEntryStruct)
		return logEntryStruct
	#



	def _descend(self, logEntryStruct):
		nextID = logEntryStruct[1]
		newList = logEntryStruct[7]
		return BufferLogger(self._idCounter, nextID, self._indentationLevel + 1, newList)
	#



	#
	# Forward the log data stored in this logger to another logger.
	#
	# @param		AbstractLogger logger			Another logger that will receive the log data.
	# @param		bool bClear						Clear buffer after forwarding all log data.
	#
	def forwardTo(self, logger, bClear = False):
		assert isinstance(logger, AbstractLogger)
		logger._logiAll(self.__list, True)
		if bClear:
			self.__list = []
	#



	def forwardToDescended(self, logger, text:str, bClear = False):
		self.forwardTo(logger.descend(text), bClear)
	#



	def clear(self):
		self.__list = []
	#



	def getDataAsJSON(self):
		return self.__getJSONData(self.__list)
	#



	def __getJSONData(self, items):
		ret = []
		for item in items:
			item2 = list(item)
			item2[5] = int(item2[5])
			if item2[0] == "txt":
				pass
			elif item2[0] == "ex":
				pass
			elif item2[0] == "desc":
				item2[7] = self.__getJSONData(item2[7])
			else:
				raise Exception("Implementation Error!")
			ret.append(item2)
		return ret
	#



	def getDataAsPrettyJSON(self):
		return self.__getPrettyJSONData(self.__list)
	#



	def __stackTraceElementToPrettyJSONData(self, stackTraceItem):
		return {
			"file": stackTraceItem[0],
			"line": stackTraceItem[1],
			"module": stackTraceItem[2],
			"sourcecode": stackTraceItem[3],
		}
	#

	def __getPrettyJSONData(self, items):
		ret = []
		for item in items:
			item2 = list(item)
			t = datetime.datetime.fromtimestamp(item2[4])
			jsonLogEntry = {
				"type": item2[0],
				"id": item2[1],
				"indent": item2[2],
				"timestamp": {
					"t": item2[4],
					"year": t.year,
					"month": t.month,
					"day": t.day,
					"hour": t.hour,
					"minute": t.minute,
					"second": t.second,
					"ms": t.microsecond // 1000,
					"us": t.microsecond % 1000,
				},
				"loglevel": str(item2[5]),
				"logleveln": int(item2[5]),
			}
			if item2[0] == "txt":
				jsonLogEntry["text"] = item2[6]
			elif item2[0] == "ex":
				jsonLogEntry["exception"] = item2[6]
				jsonLogEntry["text"] = item2[7]
				jsonLogEntry["stacktrace"] = [ self.__stackTraceElementToPrettyJSONData(x) for x in item2[8] ] if item2[8] else None
			elif item2[0] == "desc":
				jsonLogEntry["text"] = item2[6]
				jsonLogEntry["children"] = self.__getPrettyJSONData(item2[7])
			else:
				raise Exception("Implementation Error!")
			ret.append(jsonLogEntry)
		return ret
	#



	def __str__(self):
		return "<BufferLogger(" + hex(id(self)) + ", indent=" + str(self._indentationLevel) + ",parentID=" + str(self._parentLogEntryID) + ")>"
	#



	def __repr__(self):
		return "<BufferLogger(" + hex(id(self)) + ", indent=" + str(self._indentationLevel) + ",parentID=" + str(self._parentLogEntryID) + ")>"
	#



#









#!/usr/bin/env python3



import os
import time
import traceback
import sys
import abc

import sh

from .EnumLogLevel import *
from .AbstractLogger import *



class IntContainer(object):

	def __init__(self, initialValue = 0):
		self.value = initialValue
	#

#






#
# This logger keeps track of how many log messages of what type have been issued.
#
class DetectionLogger(AbstractLogger):



	def __init__(self, logger, logLevelCounterMap:dict = None, maxLogLevelSeen:IntContainer = None):
		super().__init__(None)

		self.__logger = logger
		if logLevelCounterMap is None:
			self.__logLevelCounterMap = {}
		else:
			assert isinstance(logLevelCounterMap, dict)
			self.__logLevelCounterMap = logLevelCounterMap

		if maxLogLevelSeen is None:
			self.__maxLogLevelSeen = IntContainer()
		else:
			assert isinstance(maxLogLevelSeen, IntContainer)
			self.__maxLogLevelSeen = maxLogLevelSeen
	#



	@staticmethod
	def create(logger):
		assert isinstance(logger, AbstractLogger)
		return DetectionLogger(logger)
	#



	def _descend(self, logEntryStruct):
		descendedLogger = self.__logger._descend(logEntryStruct)
		return DetectionLogger(descendedLogger, self.__logLevelCounterMap, self.__maxLogLevelSeen)
	#



	def _logi(self, logEntryStruct, bNeedsIndentationLevelAdaption):
		nLogLevel = int(logEntryStruct[5])
		self.__logLevelCounterMap[nLogLevel] = self.__logLevelCounterMap.get(nLogLevel, 0) + 1
		if nLogLevel > self.__maxLogLevelSeen.value:
			self.__maxLogLevelSeen.value = nLogLevel
		self.__logger._logi(logEntryStruct, bNeedsIndentationLevelAdaption)
	#



	#
	# Returns the number of log messages issued.
	#
	def getLogMsgCount(self, logLevel):
		return self.__logLevelCounterMap.get(int(logLevel), 0)
	#



	#
	# Returns the number of log messages issued.
	#
	def getLogMsgCountsIntMap(self):
		return {
			int(EnumLogLevel.TRACE) : self.__logLevelCounterMap.get(int(EnumLogLevel.TRACE), 0),
			int(EnumLogLevel.DEBUG) : self.__logLevelCounterMap.get(int(EnumLogLevel.DEBUG), 0),
			int(EnumLogLevel.NOTICE) : self.__logLevelCounterMap.get(int(EnumLogLevel.NOTICE), 0),
			int(EnumLogLevel.INFO) : self.__logLevelCounterMap.get(int(EnumLogLevel.INFO), 0),
			int(EnumLogLevel.STDOUT) : self.__logLevelCounterMap.get(int(EnumLogLevel.STDOUT), 0),
			int(EnumLogLevel.SUCCESS) : self.__logLevelCounterMap.get(int(EnumLogLevel.SUCCESS), 0),
			int(EnumLogLevel.WARNING) : self.__logLevelCounterMap.get(int(EnumLogLevel.WARNING), 0),
			int(EnumLogLevel.ERROR) : self.__logLevelCounterMap.get(int(EnumLogLevel.ERROR), 0),
			int(EnumLogLevel.STDERR) : self.__logLevelCounterMap.get(int(EnumLogLevel.STDERR), 0),
			int(EnumLogLevel.EXCEPTION) : self.__logLevelCounterMap.get(int(EnumLogLevel.EXCEPTION), 0),
		}
	#



	#
	# Returns the number of log messages issued.
	#
	def getLogMsgCountsStrMap(self):
		return {
			str(EnumLogLevel.TRACE) : self.__logLevelCounterMap.get(int(EnumLogLevel.TRACE), 0),
			str(EnumLogLevel.DEBUG) : self.__logLevelCounterMap.get(int(EnumLogLevel.DEBUG), 0),
			str(EnumLogLevel.NOTICE) : self.__logLevelCounterMap.get(int(EnumLogLevel.NOTICE), 0),
			str(EnumLogLevel.INFO) : self.__logLevelCounterMap.get(int(EnumLogLevel.INFO), 0),
			str(EnumLogLevel.STDOUT) : self.__logLevelCounterMap.get(int(EnumLogLevel.STDOUT), 0),
			str(EnumLogLevel.SUCCESS) : self.__logLevelCounterMap.get(int(EnumLogLevel.SUCCESS), 0),
			str(EnumLogLevel.WARNING) : self.__logLevelCounterMap.get(int(EnumLogLevel.WARNING), 0),
			str(EnumLogLevel.ERROR) : self.__logLevelCounterMap.get(int(EnumLogLevel.ERROR), 0),
			str(EnumLogLevel.STDERR) : self.__logLevelCounterMap.get(int(EnumLogLevel.STDERR), 0),
			str(EnumLogLevel.EXCEPTION) : self.__logLevelCounterMap.get(int(EnumLogLevel.EXCEPTION), 0),
		}
	#



	#
	# Indicates if this logger has seen such a log message.
	#
	def hasLogMsg(self, logLevel):
		return self.__logLevelCounterMap.get(int(logLevel), 0) > 0
	#



	def hasAtLeastWarning(self):
		return self.__maxLogLevelSeen.value >= int(EnumLogLevel.WARNING)
	#

	def hasAtLeastError(self):
		return self.__maxLogLevelSeen.value >= int(EnumLogLevel.ERROR)
	#

	def hasAtLeastException(self):
		return self.__maxLogLevelSeen.value >= int(EnumLogLevel.EXCEPTION)
	#

	def hasException(self):
		return self.hasLogMsg(EnumLogLevel.EXCEPTION)
	#

	def hasStdErr(self):
		return self.hasLogMsg(EnumLogLevel.STDERR)
	#

	def hasError(self):
		return self.hasLogMsg(EnumLogLevel.ERROR)
	#

	def hasStdOut(self):
		return self.hasLogMsg(EnumLogLevel.STDOUT)
	#

	def hasWarning(self):
		return self.hasLogMsg(EnumLogLevel.WARNING)
	#

	def hasInfo(self):
		return self.hasLogMsg(EnumLogLevel.INFO)
	#

	def hasNotice(self):
		return self.hasLogMsg(EnumLogLevel.NOTICE)
	#

	def hasDebug(self):
		return self.hasLogMsg(EnumLogLevel.DEBUG)
	#



	def descend(self, text):
		return DetectionLogger(self.__logger.descend(text), self.__logLevelCounterMap, self.__maxLogLevelSeen)
	#



	def clear(self):
		self.__logLevelCounterMap = {}
		self.__maxLogLevelSeen.value = 0
		self.__logger.clear()
	#



#





#!/usr/bin/env python3


import datetime
import abc


from .EnumLogLevel import EnumLogLevel






def _getLogLevelStr(logLevel):
	s = str(logLevel)
	pos = s.rfind(".")
	if pos >= 0:
		s = s[pos+1:]
	return s
#



def createLogMsgTypeStrMap(bPrefixWithSpacesToSameLength = False):
	maxLogLevelLength = len("STACKTRACE")
	for logLevel in EnumLogLevel:
		s = _getLogLevelStr(logLevel)
		logLevelLength = len(s)
		if logLevelLength > maxLogLevelLength:
			maxLogLevelLength = logLevelLength

	logLevelToStrDict = {}
	for logLevel in EnumLogLevel:
		s = _getLogLevelStr(logLevel)
		if bPrefixWithSpacesToSameLength:
			while len(s) < maxLogLevelLength:
				s = " " + s
		logLevelToStrDict[logLevel] = s

	return logLevelToStrDict
#







class AbstractLogMessageFormatter(object):

	LOG_LEVEL_TO_STR_MAP = createLogMsgTypeStrMap(True)



	#
	# Create and return a string representation of the specified log entry. Overwrite this method to implement a log message formatter.
	#
	# @param		list logEntryStruct			A log entry structure. See <c>AbstractLogger._logi()</c> for a detailed description.
	# @return		str							Returns the string representation of the log message.
	#
	@abc.abstractmethod
	def format(self, logEntryStruct):
		pass
	#

#







#!/usr/bin/env python3



import os
import time
import traceback
import sys
import abc
import datetime
import re

import jk_exceptionhelper

from .EnumLogLevel import *
from .IDCounter import IDCounter






def _getLogLevelStrMap(bPrefixWithSpacesToSameLength = False):
	maxLogLevelLength = len("STACKTRACE")
	for logLevel in EnumLogLevel:
		s = _getLogLevelStr(logLevel)
		if len(s) > maxLogLevelLength:
			maxLogLevelLength = len(s)
	logLevelToStrDict = {}
	for logLevel in EnumLogLevel:
		s = _getLogLevelStr(logLevel)
		if bPrefixWithSpacesToSameLength:
			while len(s) < maxLogLevelLength:
				s = " " + s
		logLevelToStrDict[logLevel] = s
	return logLevelToStrDict
#





def _getLogLevelStr(logLevel):
	s = str(logLevel)
	pos = s.rfind(".")
	if pos >= 0:
		s = s[pos+1:]
	return s
#










class _ExceptionInChildContextException(Exception):
	pass
#



class AbstractLogger(object):
	__metaclass__ = abc.ABCMeta



	_EINSTANCE = _ExceptionInChildContextException()

	_logLevelToStrDict = _getLogLevelStrMap(True)



	def __init__(self, idCounter):
		if idCounter is None:
			idCounter = IDCounter()
		else:
			assert isinstance(idCounter, IDCounter)
		self._idCounter = idCounter
		self._indentationLevel = 0
		self._parentLogEntryID = None
	#



	#
	# Create a log entry data record.
	#
	# @param		int logEntryID			The ID of this log entry or <c>None</c> if not applicable.
	# @param		int indentation			The current indentation level. Top level entries will have a value of zero here.
	#										This value is always zero if the feature is not supported.
	# @param		int parentLogEntryID	A log entry ID of the parent of this log entry or <c>None</c> if not applicable (or not supported).
	# @param		EnumLogLevel logLevel	A singleton that represents the log level.
	# @param		str text				A text to associate with this log entry.
	# @param		list newList			Typically specify an empty list here that will hold child log entries in the future.
	#
	def _createDescendLogEntryStruct(self, logEntryID, indentation, parentLogEntryID, logLevel, text, newList):
		timeStamp = time.time()

		return ("desc", logEntryID, indentation, parentLogEntryID, timeStamp, logLevel, text, newList)
	#



	@staticmethod
	def exceptionToJSON(ex):
		exceptionObject = jk_exceptionhelper.analyseException(ex)

		return {
			"exClass": exceptionObject.exceptionClassName,
			"exText": exceptionObject.exceptionTextHR,
			"exStack": [ [x.filePath, x.lineNo, x.callingScope, x.sourceCodeLine] for x in exceptionObject.stackTrace ],
		}
	#



	#
	# Creates a log entry data record.
	#
	# @param		int logEntryID			The ID of this log entry or <c>None</c> if not applicable.
	# @param		int indentation			The current indentation level. Top level entries will have a value of zero here.
	#										This value is always zero if the feature is not supported.
	# @param		int parentLogEntryID	A log entry ID of the parent of this log entry or <c>None</c> if not applicable (or not supported).
	# @param		EnumLogLevel logLevel	A singleton that represents the log level.
	# @param		obj textOrException		Either an exception object or a text
	#
	def _createNormalLogEntryStruct(self, logEntryID, indentation, parentLogEntryID, logLevel, textOrException):
		timeStamp = time.time()

		if isinstance(textOrException, Exception):
			exceptionObject = jk_exceptionhelper.analyseException(textOrException)
			return (
				"ex",
				logEntryID,
				indentation,
				parentLogEntryID,
				timeStamp,
				logLevel,
				exceptionObject.exceptionClassName,
				exceptionObject.exceptionTextHR,
				[ (x.filePath, x.lineNo, x.callingScope, x.sourceCodeLine) for x in exceptionObject.stackTrace ]
			)

		return (
			"txt",
			logEntryID,
			indentation,
			parentLogEntryID,
			timeStamp,
			logLevel,
			textOrException.rstrip('\n')
		)
	#



	#
	# Perform logging with log level ERROR.
	#
	# @param	string text		The text to write to this logger.
	#
	def error(self, text):
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.ERROR, text), False)
	#



	#
	# Perform logging with log level EXCEPTION.
	#
	# @param	Exception exception		The exception to write to this logger.
	#
	def exception(self, exception):
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.EXCEPTION, exception), False)
	#

	#
	# Perform logging with log level ERROR.
	# This method is intended to be used in conjunction with STDERR handlers.
	#
	# @param	string text		The text to write to this logger.
	#
	def stderr(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.STDERR, text), False)
	#

	#
	# Perform logging with log level STDOUT.
	# This method is intended to be used in conjunction with STDOUT handlers.
	#
	# @param	string text		The text to write to this logger.
	#
	def stdout(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.STDOUT, text), False)
	#

	#
	# Perform logging with log level SUCCESS.
	#
	# @param	string text		The text to write to this logger.
	#
	def success(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.SUCCESS, text), False)
	#

	#
	# Perform logging with log level WARNING. This method is provided for convenience and is identical with <c>warn()</c>.
	#
	# @param	string text		The text to write to this logger.
	#
	def warning(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.WARNING, text), False)
	#

	#
	# Perform logging with log level WARNING. This method is provided for convenience and is identical with <c>warning()</c>.
	#
	# @param	string text		The text to write to this logger.
	#
	def warn(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.WARNING, text), False)
	#

	#
	# Perform logging with log level INFO.
	#
	# @param	string text		The text to write to this logger.
	#
	def info(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.INFO, text), False)
	#

	#
	# Perform logging with log level NOTICE.
	#
	# @param	string text		The text to write to this logger.
	#
	def notice(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.NOTICE, text), False)
	#

	#
	# Perform logging with log level DEBUG.
	#
	# @param	string text		The text to write to this logger.
	#
	def debug(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.DEBUG, text), False)
	#



	#
	# Perform logging with log level TRACE.
	#
	# @param	string text		The text to write to this logger.
	#
	def trace(self, text):
		text = text.rstrip('\n')
		self._logi(self._createNormalLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.TRACE, text), False)
	#



	#
	# Create a nested logger. This new logger can than be used like the current logger, but all log messages will be delivered
	# to an subordinate log structure (if supported by this logger).
	#
	def descend(self, text):
		logEntryStruct = self._createDescendLogEntryStruct(self._idCounter.next(), self._indentationLevel, self._parentLogEntryID, EnumLogLevel.INFO, text, [])
		self._logi(logEntryStruct, False)
		return self._descend(logEntryStruct)
	#



	#
	# If this logger is buffering log messages, clear all log messages from this buffer.
	# If this logger has references to other loggers, such as a <c>FilterLogger</c>
	# or a <c>MulticastLogger</c>
	#
	def clear(self):
		pass
	#



	#
	# Close this logger. Some logger make use of additional resources (such as files) which will be (permanently) closed by invoking this method.
	# By default this method does nothing. Some loggers may overwrite this method in order to make use of that functionality. After closing
	# a logger you should not invoke any more logging methods of that logger. Loggers that make use of <c>close()</c> should reject any logging
	# request after a <c>close()</c> has been invoked. <c>close()</c> must always be implemented as an indempotent operation: Redundant calls to <c>close()</c>
	# should cause no problems.
	#
	def close(self):
		pass
	#



	#
	# Perform a descending operation. Overwrite this method in subclasses.
	#
	# @param		list logEntryStruct			A log entry structure. See <c>_logi()</c> for a detailed description.
	# @return		AbstractLogger				Return a logger instance representing the logger for a descended level.
	#
	@abc.abstractmethod
	def _descend(self, logEntryStruct):
		raise NotImplementedError('subclasses must override _descend()!')
	#



	#
	# Perform the log operation. This is the heart of each logger. Overwrite this method in subclasses.
	#
	# @param		list logEntryStruct						A log entry structure. Each entry consists of the following elements:
	#														0) <c>str</c> ---- The type of the log entry: "txt", "ex", "desc"
	#														1) <c>str</c> ---- The ID of the log entry or <c>None</c> if unused
	#														2) <c>int</c> ---- The indentation level
	#														3) <c>str</c> ---- The ID of the parent log entry or <c>None</c> if unused
	#														4) <c>float</c> ---- The time stamp in seconds since epoch
	#														5) <c>EnumLogLevel</c> ---- The type of the log entry
	#														If the log entry is a text entry:
	#														6) <c>str</c> ---- The text of the log message
	#														If the log entry is a descending entry:
	#														6) <c>str</c> ---- The text of the log message
	#														7) <c>list</c> ---- A list containing nested log items
	#														If the log entry is an exception:
	#														6) <c>str</c> ---- The exception class name
	#														7) <c>str</c> ---- The exception text
	#														8) <c>list</c> ---- A list of stack trace elements or <c>None</c>
	#														Each stack trace element has the following structure:
	#														0) <c>str</c> ---- The source code file path
	#														1) <c>int</c> ---- The source code line number
	#														2) <c>str</c> ---- The source code module name
	#														3) <c>str</c> ---- The source code line in plain text where the error occurred
	# @param		bool bNeedsIndentationLevelAdaption		If <c>True</c> is specified the log entry record still needs adaption at the
	#														indentation level field. This is because it was generated somewhere else and
	#														therefor has been provided by a different logger in a maybe different indentation
	#														context.
	#
	@abc.abstractmethod
	def _logi(self, logEntryStruct, bNeedsIndentationLevelAdaption):
		raise NotImplementedError('subclasses must override _logi()!')
	#



	#
	# This method is invoked in order to log a list of log entries. After adapting the indentation level to the indentation level
	# currently used by this logger either <c>_logi()</c> is called or <c>_descend()</c> in order to perform the logging.
	#
	# The default implementation provided here will perform indentation level adaption as needed. In order to do so a copy of the
	# raw log entry is created.
	#
	# @param		list logEntryStruct						A log entry structure. See <c>_logi()</c> for a detailed description.
	# @param		bool bNeedsIndentationLevelAdaption		If <c>True</c> is specified the log entry records still needs adaption at the
	#														indentation level field. This is because it was generated somewhere else and
	#														therefor has been provided by a different logger in a maybe different indentation
	#														context.
	#
	def _logiAll(self, logEntryStructList, bNeedsIndentationLevelAdaption):
		for logEntryStruct in logEntryStructList:
			#logEntryStruct = list(logEntryStruct)
			#logEntryStruct[2] = self._indentationLevel
			self._logi(logEntryStruct, bNeedsIndentationLevelAdaption)
			if logEntryStruct[0] == "desc":
				logEntryStructClone = (
					logEntryStruct[0],
					logEntryStruct[1],
					logEntryStruct[2],
					logEntryStruct[3],
					logEntryStruct[4],
					logEntryStruct[5],
					logEntryStruct[6],
					[]
				)
				self._descend(logEntryStructClone)._logiAll(logEntryStruct[7], bNeedsIndentationLevelAdaption)
	#



	def __enter__(self):
		return self
	#



	def __exit__(self, etype, value, traceback):
		if etype != None:
			if etype == _ExceptionInChildContextException:
				raise AbstractLogger._EINSTANCE
			e = etype(value)
			self.exception(e)
			raise AbstractLogger._EINSTANCE
		return False
	#



#




#!/usr/bin/env python3


import datetime

from enum import Enum

from .EnumLogLevel import EnumLogLevel
from .AbstractLogMessageFormatter import AbstractLogMessageFormatter





#
# This is a default formatter for log messages. It produces human readable output for log messages.
#
class ColoredLogMessageFormatter(AbstractLogMessageFormatter):

	class EnumOutputMode(Enum):
		VERY_SHORT = 0
		SHORTED = 10
		FULL = 20
	#


	#
	# Color codes:
	#
	# * dark foregrounds
	#	* 30 = black
	#	* 31 = dark red
	#	* 32 = dark green
	#	* 33 = brown
	# 	* 34 = dark blue
	# 	* 35 = dark pink
	# 	* 36 = dark cyan
	#	* 37 = dark grey
	#
	# * normal foregrounds
	# 	* 90 = dark grey
	# 	* 91 = bright red
	# 	* 92 = green
	# 	* 93 = yellow
	# 	* 94 = blue
	# 	* 95 = pink
	# 	* 96 = cyan
	# 	* 97 = white
	#
	# * backgrounds:
	# 	* 100 = grey
	# 	* 101 = orange
	# 	* 102 = light green
	# 	* 103 = yellow
	# 	* 104 = light blue
	# 	* 105 = light pink
	# 	* 106 = light cyan
	# 	* 107 = white
	# 


	LOG_LEVEL_TO_COLOR_MAP = {
		EnumLogLevel.TRACE: '\033[90m',
		EnumLogLevel.DEBUG: '\033[90m',
		EnumLogLevel.NOTICE: '\033[90m',
		EnumLogLevel.STDOUT: '\033[97m',
		EnumLogLevel.INFO: '\033[37m',
		EnumLogLevel.WARNING: '\033[93m',
		EnumLogLevel.ERROR: '\033[91m',
		EnumLogLevel.STDERR: '\033[91m',
		EnumLogLevel.EXCEPTION: '\033[91m',
		EnumLogLevel.SUCCESS: '\033[92m',
	}
	#STACKTRACE_COLOR = "\033[38;2;204;102;0m"
	#STACKTRACE_COLOR = "\033[93m"
	STACKTRACE_COLOR = "\033[31m"
	RESET_COLOR = '\033[0m'



	def __init__(self, bIncludeIDs = False, fillChar = "\t"):
		self.__fillChar = fillChar
		self.__indentBuffer = fillChar
		self.__includeIDs = bIncludeIDs
		self.__outputMode = ColoredLogMessageFormatter.EnumOutputMode.FULL
	#



	def setOutputMode(self, outputMode:EnumOutputMode):
		if outputMode is None:
			outputMode = ColoredLogMessageFormatter.EnumOutputMode.FULL
		self.__outputMode = outputMode
	#



	#
	# Create and return a string representation of the specified log entry.
	#
	# @param		list logEntryStruct			A log entry structure. See <c>AbstractLogger._logi()</c> for a detailed description.
	# @return		str							Returns the string representation of the log message.
	#
	def format(self, logEntryStruct):
		sID = str(logEntryStruct[1]) if (logEntryStruct != None) else "-"
		indentationLevel = logEntryStruct[2]
		while indentationLevel > len(self.__indentBuffer):
			self.__indentBuffer += self.__fillChar
		sIndent = self.__indentBuffer[0:indentationLevel]
		sParentID = str(logEntryStruct[3]) if (logEntryStruct != None) else "-"
		sTimeStamp = "[" + datetime.datetime.fromtimestamp(logEntryStruct[4]).strftime('%Y-%m-%d %H:%M:%S') + "]"
		sLogType = AbstractLogMessageFormatter.LOG_LEVEL_TO_STR_MAP[logEntryStruct[5]]

		if self.__includeIDs:
			s3 = "(" + sParentID + "|" + sID + ") " + sTimeStamp + " "
		else:
			s3 = sTimeStamp + " "
		s1 = sIndent + ColoredLogMessageFormatter.LOG_LEVEL_TO_COLOR_MAP[logEntryStruct[5]] + s3
		s2 = sIndent + ColoredLogMessageFormatter.STACKTRACE_COLOR + s3

		if logEntryStruct[0] == "txt":
			sLogMsg = logEntryStruct[6]
			return s1 + sLogType + ": " + sLogMsg + ColoredLogMessageFormatter.RESET_COLOR
		elif logEntryStruct[0] == "ex":
			sExClass = logEntryStruct[6]
			sLogMsg = logEntryStruct[7]
			ret = []
			if logEntryStruct[8] != None:
				if self.__outputMode == ColoredLogMessageFormatter.EnumOutputMode.FULL:
					for (stPath, stLineNo, stModuleName, stLine) in logEntryStruct[8]:
						ret.append(s2 + "STACKTRACE: " + stPath + ":" + str(stLineNo) + " " + stModuleName + "    # " + stLine + ColoredLogMessageFormatter.RESET_COLOR)
				elif self.__outputMode == ColoredLogMessageFormatter.EnumOutputMode.SHORTED:
					stPath, stLineNo, stModuleName, stLine = logEntryStruct[8][-1]
					ret.append(s2 + "STACKTRACE: " + stPath + ":" + str(stLineNo) + " " + stModuleName + "    # " + stLine + ColoredLogMessageFormatter.RESET_COLOR)
			ret.append(s1 + sLogType + ": " + sExClass + ": " + sLogMsg + ColoredLogMessageFormatter.RESET_COLOR)
			return ret
		elif logEntryStruct[0] == "desc":
			sLogMsg = logEntryStruct[6]
			return s1 + sLogType + ": " + sLogMsg + ColoredLogMessageFormatter.RESET_COLOR
		else:
			raise Exception()
	#

#





COLOR_LOG_MESSAGE_FORMATTER = ColoredLogMessageFormatter()













from .IDCounter import IDCounter
from .RollOverLogFile import RollOverLogFile

from .EnumLogLevel import EnumLogLevel
from .AbstractLogger import AbstractLogger
from .BufferLogger import BufferLogger
from .ConsoleLogger import ConsoleLogger
from .DetectionLogger import DetectionLogger
from .FilterLogger import FilterLogger
from .MulticastLogger import MulticastLogger
from .NamedMulticastLogger import NamedMulticastLogger
from .NullLogger import NullLogger
#from .SimpleFileLogger import SimpleFileLogger
from .FileLogger import FileLogger
from .StringListLogger import StringListLogger
from .JSONLogger import JSONLogger

from .AbstractLogMessageFormatter import AbstractLogMessageFormatter
from .LogMessageFormatter import LogMessageFormatter, DEFAULT_LOG_MESSAGE_FORMATTER
from .ColoredLogMessageFormatter import ColoredLogMessageFormatter, COLOR_LOG_MESSAGE_FORMATTER
from .HTMLLogMessageFormatter import HTMLLogMessageFormatter, HTML_LOG_MESSAGE_FORMATTER

from .LoggerInstanceManager import LoggerInstanceManager


from ._inst import instantiateLogMsgFormatter, instantiate





__version__ = "0.2020.1.31"


#!/usr/bin/env python3



import os
import time
import traceback
import sys
import abc

import sh

from .EnumLogLevel import *
from .AbstractLogger import *





#
# This logger will broadcast log messages to additional loggers.
#
class NamedMulticastLogger(AbstractLogger):



	def __init__(self, loggerMap = None, idCounter = None, indentationLevel = 0, parentLogEntryID = 0):
		super().__init__(idCounter)
		self._indentationLevel = indentationLevel
		self._parentLogEntryID = parentLogEntryID

		if loggerMap is not None:
			assert isinstance(loggerMap, dict)
			for k, v in loggerMap.items():
				assert isinstance(k, str)
				assert isinstance(v, AbstractLogger)
			self.__loggerMap = loggerMap
		else:
			self.__loggerMap = {}
	#



	@property
	def loggers(self) -> dict:
		return dict(self.__loggerMap)
	#



	def getLogger(self, loggerName) -> AbstractLogger:
		return self.__loggerMap.get(loggerName)
	#



	@staticmethod
	def create(**kwargs):
		return NamedMulticastLogger(loggerMap = kwargs)
	#



	def addLogger(self, loggerName:str, logger):
		assert isinstance(loggerName, str)
		assert isinstance(logger, AbstractLogger)
		if self.__loggerMap.get(loggerName, None) is not None:
			del self.__loggerMap[loggerName]
		self.__loggerMap[loggerName] = logger
	#



	def removeLogger(self, loggerName:str):
		assert isinstance(loggerName, str)
		if self.__loggerMap.get(loggerName, None) is not None:
			del self.__loggerMap[loggerName]
	#



	def removeAllLoggers(self):
		self.__loggerMap = {}
	#



	def _logi(self, logEntryStruct, bNeedsIndentationLevelAdaption):
		for logger in self.__loggerMap.values():
			logger._logi(logEntryStruct, True)
	#



	def _descend(self, logEntryStruct):
		nextID = logEntryStruct[1]
		newMap = {}
		for loggerName in self.__loggerMap.keys():
			logger = self.__loggerMap[loggerName]
			newMap[loggerName] = logger._descend(logEntryStruct)
		return NamedMulticastLogger(newMap, self._idCounter, self._indentationLevel + 1, nextID)
	#



	def clear(self):
		for logger in self.__loggerMap.values():
			logger.clear()
	#



	def close(self):
		for logger in self.__loggerMap.values():
			logger.close()
	#


#







#!/usr/bin/env python3



import os
import json
import time
import traceback
import sys
import abc
import datetime

import sh

from .EnumLogLevel import *
from .AbstractLogger import *
from .BufferLogger import BufferLogger





#
# This logger will buffer log messages in an internal array. Later this data can be forwarded to
# other loggers, f.e. in order to store them on disk.
#
class JSONLogger(BufferLogger):



	def __init__(self, idCounter = None, parentID = None, indentLevel = 0, logItemList = None, rootParent = None, filePath = None):
		super().__init__(idCounter, parentID, indentLevel, logItemList)

		if rootParent is not None:
			assert isinstance(rootParent, JSONLogger)
		assert isinstance(filePath, str)

		self.__filePath = filePath
		self.__filePathTmp = filePath + ".tmp"

		self.__rootParent = rootParent
	#



	@staticmethod
	def __convertRawLogData(items):
		ret = []
		for item in items:
			item = list(item)
			item[5] = EnumLogLevel.parse(item[5])
			if item[0] == "txt":
				pass
			elif item[0] == "ex":
				pass
			elif item[0] == "desc":
				item[7] = JSONLogger.__convertRawLogData(item[7])
			else:
				raise Exception("Implementation Error!")
			ret.append(item)
		return ret
	#



	@staticmethod
	def create(filePath:str):
		assert isinstance(filePath, str)

		if os.path.isfile(filePath):
			with open(filePath, "r") as f:
				jsonRawData = json.load(f)
			jsonRawData = JSONLogger.__convertRawLogData(jsonRawData)
		else:
			jsonRawData = None

		return JSONLogger(None, None, 0, jsonRawData, None, filePath)
	#



	def _logi(self, logEntryStruct, bNeedsIndentationLevelAdaption):
		super()._logi(logEntryStruct, bNeedsIndentationLevelAdaption)

		if self.__rootParent is None:
			self._saveLogData()
		else:
			self.__rootParent._saveLogData()
	#



	def _saveLogData(self):
		with open(self.__filePathTmp, "w") as f:
			json.dump(self.getDataAsJSON(), f)
		
		if os.path.isfile(self.__filePath):
			os.unlink(self.__filePath)
		os.rename(self.__filePathTmp, self.__filePath)
	#



	def _descend(self, logEntryStruct):
		nextID = logEntryStruct[1]
		newList = logEntryStruct[7]
		return JSONLogger(
			self._idCounter,
			nextID,
			self._indentationLevel + 1,
			newList,
			self.__rootParent if self.__rootParent else self,
			self.__filePath)
	#



	def __str__(self):
		return "<JSONLogger(" + hex(id(self)) + ", indent=" + str(self._indentationLevel) + ",parentID=" + str(self._parentLogEntryID) + ")>"
	#



	def __repr__(self):
		return "<JSONLogger(" + hex(id(self)) + ", indent=" + str(self._indentationLevel) + ",parentID=" + str(self._parentLogEntryID) + ")>"
	#



#









#!/usr/bin/env python3



class IDCounter(object):

	def __init__(self):
		self.__value = 0

	def next(self):
		v = self.__value
		self.__value += 1
		return v

	@property
	def value(self):
		return self.__value

	def __int__(self):
		return self.__value

	def __str__(self):
		return str(self.__value)

	def __repr__(self):
		return str(self.__value)








#!/usr/bin/env python3


import datetime


from .EnumLogLevel import EnumLogLevel
from .AbstractLogMessageFormatter import AbstractLogMessageFormatter






#
# This is a default formatter for log messages. It produces human readable output for log messages.
#
class JSONLogMessageFormatter(AbstractLogMessageFormatter):


	def __init__(self, bIncludeIDs = False, fillChar = "\t"):
		self.__fillChar = fillChar
		self.__indentBuffer = fillChar
		self.__includeIDs = bIncludeIDs



	def format(self, logEntryStruct):
		sID = str(logEntryStruct[1]) if (logEntryStruct != None) else "-"
		indentationLevel = logEntryStruct[2]
		while indentationLevel > len(self.__indentBuffer):
			self.__indentBuffer += self.__fillChar
		sIndent = self.__indentBuffer[0:indentationLevel]
		sParentID = str(logEntryStruct[3]) if (logEntryStruct != None) else "-"
		sTimeStamp = "[" + datetime.datetime.fromtimestamp(logEntryStruct[4]).strftime('%Y-%m-%d %H:%M:%S') + "]"
		sLogType = AbstractLogMessageFormatter.LOG_LEVEL_TO_STR_MAP[logEntryStruct[5]]

		s = sIndent
		if self.__includeIDs:
			s += "(" + sParentID + "|" + sID + ") "
		s += sTimeStamp + " "

		if logEntryStruct[0] == "txt":
			sLogMsg = logEntryStruct[6]
			return s + sLogType + ": " + sLogMsg
		elif logEntryStruct[0] == "ex":
			sExClass = logEntryStruct[6]
			sLogMsg = logEntryStruct[7]
			ret = []
			if logEntryStruct[8] != None:
				for (stPath, stLineNo, stModuleName, stLine) in logEntryStruct[8]:
					ret.append(s + "STACKTRACE: " + stPath + ":" + str(stLineNo) + " " + stModuleName + "    # " + stLine)
			ret.append(s + " "  + sLogType + ": " + sExClass + ": " + sLogMsg)
			return ret
		elif logEntryStruct[0] == "desc":
			sLogMsg = logEntryStruct[6]
			return s + sLogType + ": " + sLogMsg
		else:
			raise Exception()




JSON_LOG_MESSAGE_FORMATTER = JSONLogMessageFormatter()














import os
import time
import traceback
import sys
import abc
import datetime

from .EnumLogLevel import *
from .AbstractLogger import *





#
# This object represents a roll-over enabled log file on disk log messages can be written to. Data is encoded using utf-8.
#
class RollOverLogFile(object):

	#
	# Constructor method.
	#
	# @param		str filePath				The path of the file to write the log data to. If <c>rollOverMode</c> is not <c>None</c>
	#											you need to use the following shortcuts within the file path to be filled in with date and
	#											time values:
	#											* %Y - year (f.e. 2017)
	#											* %m - month (f.e. 08)
	#											* %d - day (f.e. 02)
	#											* %H - hour (f.e. 03)
	#											* %M - minute (f.e. 49)
	#											* %S - second (f.e. 07)
	#											All values used in the pattern elements are padded with zeros.
	# @param		str rollOverMode			Either specify <c>None</c> if no roll-over should be used. Otherwise
	#											specify either "second", "minute", "hour", "day" or "month".
	# @param		bool bAppendToExistingFile	If <c>True</c> a possibly already existing file will be reused, data will
	#											be appended. If <c>False</c> the file will be recreated if it already exists.
	# @param		bool bFlushAfterEveryWrite	If <c>True</c> perform a flush after every write.
	# @param		int fileMode				If not <c>None</c> <c>chmod()</c> will be called with these file modes specified here.
	#
	def __init__(self, filePath, rollOverMode, bAppendToExistingFile, bFlushAfterEveryWrite, fileMode):
		assert isinstance(filePath, str)
		assert isinstance(bFlushAfterEveryWrite, bool)
		assert isinstance(bAppendToExistingFile, bool)
		if fileMode is not None:
			assert isinstance(fileMode, int)
		if rollOverMode is not None:
			assert isinstance(rollOverMode, str)
			assert rollOverMode in ["second", "minute", "hour", "day", "month", "none"]
			if rollOverMode == "none":
				rollOverMode = None

		self.__fileCreationTimeStamp = datetime.datetime.now()
		self.__rollOverMode = rollOverMode
		self.__filePath = filePath
		self.__bFlushAfterEveryWrite = bFlushAfterEveryWrite
		self.__bAppendToExistingFile = bAppendToExistingFile
		self.__fileMode = fileMode

		realFilePath = self.__buildFilePath(filePath, self.__fileCreationTimeStamp)
		if not bAppendToExistingFile:
			if os.path.isfile(realFilePath):
				os.unlink(realFilePath)
		self.__f = open(realFilePath, "a")
		if fileMode is not None:
			os.chmod(realFilePath, fileMode)

		self.__bIsDirty = False
	#



	def __isRollOverRequired(self, now):
		if self.__rollOverMode is None:
			return False

		if self.__rollOverMode == "second":
			if (now.second != self.__fileCreationTimeStamp.second) or (now.minute != self.__fileCreationTimeStamp.minute) \
				or (now.hour != self.__fileCreationTimeStamp.hour) or (now.day != self.__fileCreationTimeStamp.day) \
				or (now.month != self.__fileCreationTimeStamp.month) or (now.year != self.__fileCreationTimeStamp.year):
				return True
		elif self.__rollOverMode == "minute":
			if (now.minute != self.__fileCreationTimeStamp.minute) \
				or (now.hour != self.__fileCreationTimeStamp.hour) or (now.day != self.__fileCreationTimeStamp.day) \
				or (now.month != self.__fileCreationTimeStamp.month) or (now.year != self.__fileCreationTimeStamp.year):
				return True
		elif self.__rollOverMode == "hour":
			if (now.hour != self.__fileCreationTimeStamp.hour) or (now.day != self.__fileCreationTimeStamp.day) \
				or (now.month != self.__fileCreationTimeStamp.month) or (now.year != self.__fileCreationTimeStamp.year):
				return True
		elif self.__rollOverMode == "day":
			if (now.day != self.__fileCreationTimeStamp.day) \
				or (now.month != self.__fileCreationTimeStamp.month) or (now.year != self.__fileCreationTimeStamp.year):
				return True
		elif self.__rollOverMode == "month":
			if (now.month != self.__fileCreationTimeStamp.month) or (now.year != self.__fileCreationTimeStamp.year):
				return True

		return False
	#



	def __toStr(self, number, size):
		s = str(number)
		while len(s) < size:
			s = "0" + s
		return s
	#



	def __buildFilePath(self, filePath, timestamp):
		dataTriples = (
			("Y", 4, timestamp.year),
			("m", 2, timestamp.month),
			("d", 2, timestamp.day),
			("H", 2, timestamp.hour),
			("M", 2, timestamp.minute),
			("S", 2, timestamp.second)
		)

		ret = ""
		bNextCharIsSymbolic = False
		for c in filePath:
			if bNextCharIsSymbolic:
				bReplaced = False
				for (key, paddingSize, value) in dataTriples:
					if key == c:
						ret += self.__toStr(value, paddingSize)
						bReplaced = True
						break
				if not bReplaced:
					ret += c
				bNextCharIsSymbolic = False
			else:
				if c == '%':
					bNextCharIsSymbolic = True
				else:
					ret += c

		return ret
	#



	#
	# Write one or more text lines to the log file.
	#
	# @param		str prefix				If not <c>None</c> this text is prepended to the log file.
	# @param		mixed lineOrLines		A single text string or a list of text lines to write. Don't specify <c>None</c> here.
	#
	def print(self, prefix, lineOrLines):
		if self.__f.closed:
			raise Exception("File already closed!")

		now = datetime.datetime.now()
		if self.__isRollOverRequired(now):
			self.__f.flush()
			self.__f.close()
			realFilePath = self.__buildFilePath(self.__filePath, now)
			if not self.__bAppendToExistingFile:
				if os.path.isfile(realFilePath):
					os.unlink(realFilePath)
			self.__f = open(realFilePath, "a")
			if self.__fileMode is not None:
				os.chmod(realFilePath, self.__fileMode)
			self.__fileCreationTimeStamp = now

		if isinstance(lineOrLines, str):
			if prefix is None:
				self.__f.write(lineOrLines + "\n")
			else:
				self.__f.write(prefix + lineOrLines + "\n")
		else:
			if prefix is None:
				for line in lineOrLines:
					self.__f.write(line + "\n")
			else:
				for line in lineOrLines:
					self.__f.write(prefix + line + "\n")

		if self.__bFlushAfterEveryWrite:
			self.__f.flush()
			self.__bIsDirty = False
		else:
			self.__bIsDirty = True
	#



	#
	# Close the file.
	#
	def close(self):
		if not self.__f.closed:
			self.__f.flush()
			self.__f.close()
	#



	#
	# Checks if the file is closed.
	#
	# @return	bool		Returns <c>true</c> if the file is already closed.
	#
	@property
	def isClosed(self):
		return self.__f.closed
	#



	#
	# Checks if the file is closed.
	#
	# @return	bool		Returns <c>true</c> if the file is already closed.
	#
	@property
	def closed(self):
		return self.__f.closed
	#



#












#!/usr/bin/env python3



import os
import time
import traceback
import sys
import abc

import sh

from .EnumLogLevel import *
from .AbstractLogger import *





#
# This logger just consumes all loggin output without doing anyting with it.
#
class NullLogger(AbstractLogger):


	def __init__(self, idCounter = None):
		super().__init__(idCounter)
	#



	@staticmethod
	def create():
		return _INSTANCE
	#



	def _logi(self, logEntryStruct, bNeedsIndentationLevelAdaption):
		pass
	#



	def _descend(self, logEntryStruct):
		return self
	#



#



_INSTANCE = NullLogger()













